;
;Copyright (C) 2005 Patricio Rojo
;
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
;

;+
; NAME:
;	DEFRINGEFLAT
;
; PURPOSE:
;	This function removes the fringes from a flat field.
;
; CATEGORY:
;	Pato's Array Data Reduction Potpourri.
;
; CALLING SEQUENCE:
;
;	defringedflat = DEFRINGEFLAT(Flat, Refper)
;
;
; INPUTS:
;    Flat:       Two dimensional array containing the flat field to
;                defringe.
;    Refper:     Approximate fringe period (in pixels) at position
;                COLREF, ROWREF.
;
; KEYWORD PARAMETERS:
;  ARRAY KEYWORDS
;    BINW:       Number of rows included in median average to improve
;                signal-to-noise ratio. Default is 10.
;    COLREF:     Reference image column for which REFPER value is
;                given. Default is middle column.
;    ROWREF:     Reference row for which REFPER value is
;                given. Default is middle row.
;    TRIM:       Ignore datapoints outside these boundaries. Format is
;                [x0, y0, x1, y1].
;
;    GOODPIX:    Mask of good pixels. Must have same shape as Flat
;                array. 1 is good pixel, 0 is bad. 
;    GPIXRAD:    Radius of median filter to apply whenever a good
;                pixel mask is supplied. The median value will replace
;                the bad value. Default is 5.
;
;  TRACE FITTING KEYWORDS
;    SILENT:     Set this for a silent run without any kind of
;                progress output. Error messages will show, though.
;    FITNAME:    What kind of fit to do. Currently 'gauss', 'gaussest',
;                and 'true' are supported. Default is 'gauss'.
;    OSAMP:      Trace-fit oversampling factor. Default is 10.
;    BACKSUB:    If set, subtract the background from the trace,
;                i.e. consider the background as part of the flat and
;                not part of the fringe pattern.
;    TRACELIM:   Specify this to have a limit on the trace
;                value. It should be a two-element array. The center
;                of the trace won't go below TRACELIM[0] or above
;                TRACELIM[1].
;    PERUPDATE:  Set this to update the value of REFPER with the
;                fitted period in the initial position. Useful when
;                processing several flats in a loop.
;    SHOWTRACE:  Show the fitted trace after having calculated it for
;                each row.
;
;  WAVELET PARAMETERS KEYWORDS
;    WAVEMOTHER: Name of wavelet to use. Currently only Morlet, Paul,
;                or DOG, with WAVEPARAM 6, 4, or 2, respectively, are
;                supported.
;    WAVEPARAM:  Parameter of the wavelet, currently only some values
;                are supported, see WAVEMOTHER.
;    NOPAD:      Set this flag not to zero-pad the data before doing
;                the wavelet transform.
;
;  FRINGE PARAMETERS KEYWORDS
;    SMOOTHFRINGE: Set this to patch and smooth (see below) the
;                parameters of the trace in the wavelet space before
;                reconstructing.
;    NSIGMA:     Rejection threshold for fitted fringe parameters. If 
;                any parameter is more than NSIGMA standard deviations
;                different from its median value then patch it. It has
;                an effect only when SMOOTHFRINGE is
;                specified. Default is 1.5.
;    RADMEDFILTER: Radius of the median filter to apply when
;                discriminating bad parameter values. It has an effect
;                only when SMOOTHFRINGE is specified. Default is 15.
;    RADSMOOTH:  Radius of mean filter to apply when smoothing in the
;                parameter space. It has an effect only when
;                SMOOTHFRINGE is specified. Default is 9.
;
;  OPTIONAL OUTPUT KEYWORD
;    FRINGE:     Returns the two dimensional fringe that was
;                constructed from the wavelet fit.
;
; OUTPUTS:
;       This function returns the corrected flat (input Flat minus the
;       fitted fringe).
;
; SIDE EFFECTS:
;    Value of input parameter Refper is overwritten when using
;    PERUPDATE option.
;
; RESTRICTIONS:
;     Do no operate heavy machinery while using this:D.
;
; PROCEDURE:
;       A 1D median average of BINW rows is computed for each of the
;     rows. Let's call each of the rows obtained in this manner
;     a meanrow.
;       For each of those meanrows, the slope is subtracted
;     through a linear fit inside the TRIM limits, and then every
;     point outside the TRIM limits is replaced with zeroes. Then, a
;     wavelet transform is performed according to the WAVEMOTHER and
;     the WAVEPARAM keywords.
;       Starting from Refper at position COLREF and ROWREF, the trace
;     of prominent periods in the absolute value of the wavelet space
;     is followed along the columns. The trace profile is fit by a
;     FITNAME function. Repeating this procedure for each row, we
;     obtain a fitted set of parameters (e.g., center, width,
;     amplitude of a Gaussian) for each pixel in the array.
;       If SMOOTHFRINGE is set, then each of the parameters is
;     patched (compared against the median value using NSIGMA and
;     RADMEDFILTER) and painted (smoothed according to RADSMOOTH).
;       Then, the wavelet transform is reconstructed from the
;     (smoothed) parameters. Finally, an inverse wavelet transform is
;     performed to obtain the fringe pattern, which is then removed
;     from the flat.
;
; DEPENDS ON:
;     meanfilter.pro  Preoduces a mean filter of a two dimensional
;                     array.
;     maxtrace.pro    Finds a trace in a 2 dimensional array.
;     waverecon.pro   Produces an inverse wavelet transform.
;     wavelet.pro     Do a forward or inverse wavelet
;                     transform.
;                     Routine by Torrence and Compo. It is
;                     available from the DEFRINGEFLAT homepage
;                     http://.
;                     The latest version should be available directly
;                     from the authors at
;                     http://paos.colorado.edu/research/wavelets/
;
;
; EXAMPLE:
;	Read a flat and defringe it. Start with a reference period of
;     40 pixels at the center of the array, no trimming. Use a
;     Gaussian fit, interpolate period in the wavelet transform to 10
;     times more points for improved trace fitting. Show the fitted
;     trace after finding it in every row. Patch and smooth each of
;     the parameters of the Gaussian fit (center, width, height, and
;     background).
;
;         flat       = readfits('Flat.fits', h)
;         dim        = size(flat, /dim)
;         ref_period = 40
;         clean = DEFRINGEFLAT(flat, ref_period, binw=20, trim=[0,0,dim],   $
;                              colref=dim[0]/2.0, rowref=dim[1]/2.0,        $
;                              fitname='gauss', osamp=10, /show, /smooth,   $
;                              radsmooth=5, radmedfilter=10, nsigma=3)
;
; MODIFICATION HISTORY:
; 	Written by:	Pato Rojo, Cornell.  Aug 2005
;			pato@astro.cornell.edu
;-

function findandpatch_defringe, data, radius, nsigma
compile_opt idl2, hidden
;look for outliers and patch

med = median(data, radius, /double)
difprm = data - med
outliers = where(abs(difprm) gt stddev(difprm)*nsigma)

if outliers[0] eq -1 then return, data

ret = data
ret[outliers] = med[outliers]

return, ret
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro pnp_defringe, prmfit, radsmooth, radmedfilter, $
                  nsigma, trim, silent
compile_opt idl2, hidden
;Patch and paint(smooth) each of the fitted parameters

if ~ silent then print, ' Correcting parameters', format='(a,$)'
for e = 0, (size(prmfit, /dim))[2]-1 do begin 
    if ~ silent then print, string(e, format='(i2)') + ".", format='(a,$)'
    ;Patch. Find and fix outliers
    smoothed = findandpatch_defringe(reform(prmfit[trim[0]:trim[2],     $
                                                   trim[1]:trim[3],e]), $
                                     radmedfilter, nsigma)
                                ;Paint. Smooth parameters
    if ~ silent then print,".", format='(a,$)'
    smoothed = meanfilter(smoothed, radsmooth)
    if ~ silent then print, ".", format='(a,$)'

    prmfit[trim[0]:trim[2], trim[1]:trim[3], e] = smoothed
endfor
if ~ silent then print, " "

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

function maketrace_defringe, prm, idx, period, fitname, trim, noval
compile_opt idl2, hidden
;Rebuild trace

nwa = dblarr(n_elements(idx), n_elements(period))
cnt = prm[*,1]
nprm = (size(prm, /dim))[1]

for xi = trim[0], trim[2] do if (cnt[xi] ne noval) then break

switch fitname of
    'gauss': 
    'gaussest': begin
        for i = xi, trim[2] do begin
            if (cnt[i] eq noval) then break
            nwa[i,*] = prm[i,0] * exp(-((period-prm[i,1])/prm[i,2])^2) + $
              ((nprm eq 4)?prm[i,3]:0.0)
        endfor
        break
    end
    else: message, "Method '" + fitname + $
      "' not enabled in maketrace_defringe"
endswitch

return, nwa

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro buildfringe_defringe, silent, prmfit, nwp, trim, noval, $ ;border
                          cdelta, psi0, period, scale, fitname, $ ;parameters
                          fringe ;returns
compile_opt idl2, hidden
;rebuild fringe from fitted wavelet

nx = (size(prmfit, /dim))[0]
ny = (size(prmfit, /dim))[1]
idx = indgen(nx)

if size(flat, /type) eq 5 then fringe = dblarr(nx, ny) $
else fringe = fltarr(nx, ny)

;reconstruct
if ~ silent then print," Reconstructing..."
for j = trim[1], trim[3] do begin
    ;rebuild trace
    nwa = maketrace_defringe(reform(prmfit[*,j,*]), idx, period, $
                             fitname, trim, noval)

    cnw = complex(nwa*cos(nwp[*,*,j]), nwa*sin(nwp[*,*,j]))

    ;rebuild row
    fringe[*,j] = waverecon(cnw, 1, scale, cdelta, psi0)
endfor

fringe[0:(trim[0]-1) > 0, *] = 0
fringe[(trim[2]+1) < (nx-1):*, *] = 0

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro tracefit_defringe, flat, trim, binw, osamp, $ ;basics
                       rrefper, colref, rowref, $ ;reference
                       doshowtrace, fitname, $
                       tracelim, silent, $ ;flags
                       waveparam, wavemother, pad, noval, $ ;wavelet parameters
                       prmfit, nwp, period, scale ;returns
compile_opt idl2, hidden
;find fringe from fitted wavelet of data

if ~ silent then print,' Fitting trace for each row...'
nx = (size(flat, /dim))[0]
ny = (size(flat, /dim))[1]
p = indgen(nx)
cbin = binw/2
idx = indgen(nx)

w = wavelet(p, 1, coi=coi, scale=scale, period=period, $
            mother=wavemother, param=waveparam, pad=pad)
nwp = dblarr(nx, n_elements(scale), ny)

switch fitname of
    'gauss': if ~ silent then print, 'F', format='(a,$)'
    'gaussest': begin
        if ~ silent then print,'G:',format='(a,$)'
        prmfit = dblarr(nx, ny, 4)
        break
    end
    else: message, "Fitting method '" + fitname + "' not enabled in tracefit"
endswitch

finrow = trim[3] - binw
steprow = 1
for order = 0, 1 do begin
    refper = rrefper

    for j = rowref - cbin, finrow, steprow do begin
        ;median average, subtract slope, and nullify borders
        p = median(reform(flat[*,j:j+binw]), dimension=2, /double)
        a = linfit(idx[trim[0]:trim[2]], p[trim[0]:trim[2]])
        p = p - a[0] - a[1]*idx
        p[0:trim[0]] = 0
        p[trim[2]:*] = 0

        ;produce wavelet
        w = wavelet(p, 1, coi=coi, scale=scale, period=period, $
                    mother=wavemother, param=waveparam, pad=pad)
        if ~ silent then begin
            if steprow lt 0 then print, '<', format='(a,$)' $
            else print,'>', format='(a,$)'
        endif
        prmfit[*,j+cbin,*] = maxtrace(abs(w), period, colref, refper,     $
                                      osamp=osamp, fit=fitname,           $
                                      /expy, ignore=[trim[0],trim[2]],    $
                                      tracelim=tracelim, noval=noval)

        ;show if requested
        if doshowtrace then begin
            !p.position = [.1,.05,.95,.8]
            ttrace = where(prmfit[*,j+cbin,1] ne 1)
            mval = max(abs(w[colref-5:colref+5, refper-5:refper+5]))
            contour, abs(w), idx, period, $
              /xstyle, xtitle='Pixel', $
              /ystyle, ytitle='Period', yrange=[100,10], /ytype, $
              zrange=[1.3*mval,0], /fill, nlevels=25
            plots, idx, coi, noclip=0, thick=2
            plots, idx[ttrace], prmfit[ttrace,j+cbin,1], noclip=0
            plots, idx, replicate(40,1024), psym=0, noclip=0
            !p.position = [.1,.8,.95,.95]
            plot, p, /ystyle, /xstyle, xtickname=replicate(" ",30), $
              /noerase, title='Row ' + string(j)
            !p.position = 0
        endif

        ;store phase
        nwp[*, *, j+cbin] = atan(w, /phase)
        refper = prmfit[colref, j+cbin, 1]
    endfor

    finrow  = trim[1]
    steprow = -1
endfor
if ~ silent then print,''

end


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


pro trueshape_defringe, flat, trim, binw, osamp, rrefper, colref, rowref, $
                        waveparam, wavemother, pad, cdelta, psi0, fitname,  $
                        doshowtrace, backsub, tracelim, silent, fringe
compile_opt idl2, hidden
;true shape search of fringe's wavelet

if ~ silent then print, ' Fitting trace for each row...'
nx = (size(flat, /dim))[0]
ny = (size(flat, /dim))[1]

if size(flat, /type) eq 5 then fringe = dblarr(nx, ny) $
else fringe = fltarr(nx,ny)

cbin = binw/2
idx = indgen(nx)

finrow = trim[3] - binw
steprow = 1
if ~ silent then print,'TS:',format='(a,$)'
for order = 0, 1 do begin
    refper = rrefper

    for j = rowref - cbin, finrow, steprow do begin

        ;take medianized row, subtract slope, nullify borders
        p = median(reform(flat[*, j:j+binw]), dimension=2, /double)
        a = linfit(idx[trim[0]:trim[2]], p[trim[0]:trim[2]])
        p = p - a[0] - a[1]*idx
        p[0:trim[0]] = 0
        p[trim[2]:*] = 0

        ;produce wavelet
        w = wavelet(p, 1, coi=coi, scale=scale, period=period, $
                    mother=wavemother, param=waveparam, pad=pad)

        if ~ silent then begin
            if steprow lt 0 then print, '<', format='(a,$)' $
            else print, '>', format='(a,$)'
        endif

        nwa = maxtrace(abs(w), period, colref, refper,                  $
                       osamp=osamp, fitname=fitname,                    $
                       /expy, ignore=[trim[0],trim[2]], peaks=peaks,    $
                       trueback=trueback, tracelim=tracelim )

        ;show if requested
        if doshowtrace then begin
            !p.position = [.1,.05,.95,.8]
            mval = max(nwa[colref-5:colref+5, refper-5:refper+5])
            contour, nwa, idx, period, $
              /xstyle, xtitle='Pixel', /ystyle, ytitle='Period',   $
              yrange=[100,10], /ytype, zrange=[1.3*mval,0], /fill, $
              nlevels=25
            plots, idx, coi, noclip=0, thick=2
            plots, idx, peaks, psym=0, noclip=0
            !p.position = [.1,.8,.95,.95]
            plot, p, /ystyle, /xstyle, xtickname=replicate(" ",30), $
              /noerase, title='Row '+string(j),                     $
              yrange=mean(p)+[-1,1]*4*stddev(p)
            !p.position = 0
        endif

        nwp = atan(w, /phase)

        if backsub then begin
            for x=0, nx-1 do begin
                shape = where(nwa[x,*] gt 0)
                if shape[0] eq -1 then continue
                nwa[x,shape] = nwa[x,shape] - trueback[x,0] - $
                  period[shape]*trueback[x,1]
            endfor
        endif
        cnw = complex(nwa*cos(nwp), nwa*sin(nwp))

        ;rebuild row
        fringe[*, j+cbin] = waverecon(cnw, 1, scale, cdelta, psi0)

        ;adjust refper for the next row
        refper = (where(period gt peaks[colref]))[0]
        if period[refper]-peaks[colref] gt $
          0.5*(period[refper]-period[refper-1]) then refper -= 1
    endfor

    finrow = trim[1]
    steprow = -1
endfor
if ~ silent then print,''

end


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


function defringeflat, flat, refper,                               $
                       trim=itrim, binwidth=binw, osamp=osamp,     $
                       wavemother=wavemother, waveparam=waveparam, $
                       colref=colref, rowref=rowref,               $
                       backsub=backsub, silent=silent,             $
                       showtrace=showtrace, fitname=fitname,       $
                       smoothfringe=smoothfringe, nsigma=nsigma,   $
                       radmedfilter=radmedfilter,                  $
                       radsmooth=radsmooth, nopad=nopad,           $
                       goodpix=goodpix, gpixrad=gpixrad,           $
                       perupdate=perupdate, tracelim=tracelim,     $
                       fringe=fringe
compile_opt idl2
;Defringe a flat

nx = (size(flat, /dim))[0]
ny = (size(flat, /dim))[1]
if (size(flat, /n_dimen) ne 2) then $
  message,"First parameter of defringeflat is not a two-dimensional array"

case n_elements(itrim) of
    0: trim = [0,0,nx-1,ny-1]
    4: trim = itrim
    else: $
      message, "Trim parameter is invalid. It need to be a 4 elements array"
endcase

pad = 1
if keyword_set(nopad)       then pad = 0

if ~ keyword_set(silent)       then silent       = 0
if ~ keyword_set(nsigma)       then nsigma       = 1.5
if ~ keyword_set(radmedfilter) then radmedfilter = 15
if ~ keyword_set(radsmooth)    then radsmooth    = 9
if ~ keyword_set(backsub)      then backsub      = 0
if ~ keyword_set(showtrace)    then showtrace    = 0
if ~ keyword_set(wavemother)   then wavemother   = 'Morlet'
if ~ keyword_set(waveparam)    then waveparam    = 6
if ~ keyword_set(fitname)      then fitname      = 'gauss'
if ~ n_elements(binw)          then binw         = 10
if ~ n_elements(osamp)         then osamp        = 10
if ~ n_elements(colref)        then colref       = nx/2
if ~ n_elements(rowref)        then rowref       = ny/2

switch wavemother of
    'Morlet': begin
        if waveparam eq 6 then begin
            Cdelta = 0.776
            psi0 = !PI^(-0.25)
            break
        endif
    end
    'Paul': begin
        if waveparam eq 4 then begin
            Cdelta = 1.132
            psi0 = 2.^waveparam * FACTORIAL(waveparam) / $
              SQRT(!PI * FACTORIAL(2*waveparam))
            break
        endif
    end
    'Dog': begin
        if waveparam eq 2 then begin
            Cdelta = 3.541
            psi0 = 0.867325
            break
        endif
    end
    else: $
      message, string("Cannot do a wavelet of type " + wavemother + $
                      " (parameter ", waveparam, "), try Morlet, Paul, or" + $
                      " DOG only, with parameters 6, 4, or 2, repectively", $
                      format='(a,i0,a)')
endswitch

;if a good pixel mask is supplied, replace those values for their
;median value.
if keyword_set(goodpix) then begin
    if ~ keyword_set(gpixrad) then gpixrad = 5
    idx = where(goodpix eq 0)
    flat[idx] = (median(flat, gpixrad))[idx]
endif

if strpos(fitname, 'true') ne -1 then begin
    trueshape_defringe, flat, trim, binw, osamp, refper, colref, rowref, $
      waveparam, wavemother, pad, cdelta, psi0, fitname, showtrace,      $
      backsub, tracelim, silent, fringe
    if keyword_set(perupdate) then refper = prmfit[colref, rowref, 1]
    return, flat - fringe
endif

;lookfor trace to fit
tracefit_defringe, flat, trim, binw, osamp, refper, colref, rowref, $
  showtrace, fitname, tracelim, silent, waveparam, wavemother, pad, $
  noval, prmfit, nwp, period, scale
if keyword_set(perupdate) then refper = prmfit[colref, rowref, 1]
trim += [0, 1, 0, -1] * binw/2
if backsub then switch fitname of
    'gauss': 
    'gaussest': begin
        prmfit = prmfit[*,*,0:2]
        break
    end
    else: message, "Fitting method '" + fitname + "' not enabled in defringe"
endswitch

if keyword_set(smoothfringe) then begin
                                ;patch, paint
    pnp_defringe, prmfit, radsmooth, radmedfilter, nsigma, trim, silent
    trim += [1, 1, -1, -1] * radsmooth
endif

;build fringe pattern
buildfringe_defringe, silent, prmfit, nwp, trim, noval, cdelta, psi0, $
  period, scale, fitname, fringe

return, flat - fringe

end
